# quidditas setup.py
from distutils.core import setup

setup(
    name = "quidditas",
    packages = ["quidditas"],
    version= "0.9",
    description = "Entity framework for games",
    author = "Dominik Schacht",
    author_email = "domschacht@gmail.com",
    url = "http://quidditas.googlecode.com/",
    keywords = ["entity", "entities", "games"],
    license = "BSD",
    classifiers = [
        "License :: OSI Approved :: BSD License",
        "Intended Audience :: Developers",
        "Development Status :: 4 - Beta",
        "Topic :: Games/Entertainment",
    ]

)


